/*
    PPToolsPopupPanelController.m

    Copyright 2013-2016 Josh Freeman
    http://www.twilightedge.com

    This file is part of PikoPixel for Mac OS X and GNUstep.
    PikoPixel is a graphical application for drawing & editing pixel-art images.

    PikoPixel is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3,
    as published by the Free Software Foundation.

    PikoPixel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#import "PPToolsPopupPanelController.h"

#import "PPDocument.h"
#import "NSColor_PPUtilities.h"
#import "PPUIColors_Panels.h"
#import "PPPopupPanelActionKeys.h"


#define kToolsPopupPanelNibName  @"ToolsPopupPanel"


@interface PPToolsPopupPanelController (PrivateMethods)

- (void) handlePPDocumentNotification_ChangedActiveTool: (NSNotification *) notification;

- (void) updateToolsMatrix;
- (PPToolType) toolTypeForSelectedToolMatrixCell;

@end

@implementation PPToolsPopupPanelController

- (void) dealloc
{
    [_activeToolCellColor release];
    [_inactiveToolCellColor release];

    [super dealloc];
}

#pragma mark Actions

- (IBAction) toolsMatrixButtonPressed: (id) sender
{
    [_ppDocument setSelectedToolType: [self toolTypeForSelectedToolMatrixCell]];
}

#pragma mark NSWindowController overrides

- (void) windowDidLoad
{
    _activeToolCellColor =
        [[NSColor ppCenteredVerticalGradientPatternColorWithHeight:
                                                            [_toolsMatrix cellSize].height
                            innerColor: kUIColor_ToolsPanel_ActiveToolCellGradientInnerColor
                            outerColor: kUIColor_ToolsPanel_ActiveToolCellGradientOuterColor]
                retain];

    _inactiveToolCellColor = [kUIColor_ToolsPanel_InactiveToolCellColor retain];

    [[_toolsMatrix cells] makeObjectsPerformSelector: @selector(setBackgroundColor:)
                            withObject: _inactiveToolCellColor];

    [super windowDidLoad];
}

#pragma mark PPPopupPanelController overrides

+ (NSString *) panelNibName
{
    return kToolsPopupPanelNibName;
}

- (NSColor *) backgroundColorForPopupPanel
{
    return kUIColor_ToolsPopupPanel_Background;
}

- (void) setupPanelForCurrentPPDocument
{
    [super setupPanelForCurrentPPDocument];

    [self updateToolsMatrix];
}

- (bool) handleActionKey: (NSString *) key
{
    if ([key isEqualToString: kToolsPopupPanelActionKey_SelectLastTool])
    {
        [_ppDocument setSelectedToolTypeToLastSelectedType];

        return YES;
    }

    return NO;
}

- (void) handleDirectionCommand: (PPDirectionType) directionType
{
    NSInteger numRows, numCols, selectedRow, selectedCol;
    NSCell *selectedCell;

    [_toolsMatrix getNumberOfRows: &numRows columns: &numCols];

    selectedCell = [_toolsMatrix selectedCell];

    if (!selectedCell
        || ![_toolsMatrix getRow: &selectedRow column: &selectedCol ofCell: selectedCell])
    {
        goto ERROR;
    }

    switch (directionType)
    {
        case kPPDirectionType_Left:
        {
            if (--selectedCol < 0)
            {
                selectedCol = numCols - 1;
            }
        }
        break;

        case kPPDirectionType_Right:
        {
            if (++selectedCol >= numCols)
            {
                selectedCol = 0;
            }
        }
        break;

        case kPPDirectionType_Up:
        {
            if (--selectedRow < 0)
            {
                selectedRow = numRows - 1;
            }
        }
        break;

        case kPPDirectionType_Down:
        {
            if (++selectedRow >= numRows)
            {
                selectedRow = 0;
            }
        }
        break;

        default:
        {
            goto ERROR;
        }
        break;
    }

    [_toolsMatrix selectCellAtRow: selectedRow column: selectedCol];

    [_ppDocument setSelectedToolType: [self toolTypeForSelectedToolMatrixCell]];

    return;

ERROR:
    return;
}

#pragma mark PPDocument notifications

- (void) addAsObserverForPPDocumentNotifications
{
    if (!_ppDocument)
        return;

    [[NSNotificationCenter defaultCenter]
                                addObserver: self
                                selector:
                                    @selector(handlePPDocumentNotification_ChangedActiveTool:)
                                name: PPDocumentNotification_ChangedActiveTool
                                object: _ppDocument];
}

- (void) removeAsObserverForPPDocumentNotifications
{
    [[NSNotificationCenter defaultCenter] removeObserver: self
                                            name: PPDocumentNotification_ChangedActiveTool
                                            object: _ppDocument];
}

- (void) handlePPDocumentNotification_ChangedActiveTool: (NSNotification *) notification
{
    [self updateToolsMatrix];
}

#pragma mark Private methods

- (void) updateToolsMatrix
{
    NSButtonCell *lastActiveToolCell, *activeToolCell;

    lastActiveToolCell = [_toolsMatrix cellWithTag: _tagOfActiveToolMatrixCell];
    [lastActiveToolCell setBackgroundColor: _inactiveToolCellColor];

    _tagOfActiveToolMatrixCell = [_ppDocument activeToolType];
    activeToolCell = [_toolsMatrix cellWithTag: _tagOfActiveToolMatrixCell];
    [activeToolCell setBackgroundColor: _activeToolCellColor];

    [_toolsMatrix selectCell: activeToolCell];
}

- (PPToolType) toolTypeForSelectedToolMatrixCell
{
    return (PPToolType) [[_toolsMatrix selectedCell] tag];
}

@end
